"""
Forest of trees-based ensemble methods.

Those methods include random forests and extremely randomized trees.

The module structure is the following:

- The ``BaseForest`` base class implements a common ``fit`` method for all
  the estimators in the module. The ``fit`` method of the base ``Forest``
  class calls the ``fit`` method of each sub-estimator on random samples
  (with replacement, a.k.a. bootstrap) of the training set.

  The init of the sub-estimator is further delegated to the
  ``BaseEnsemble`` constructor.

- The ``ForestClassifier`` and ``ForestRegressor`` base classes further
  implement the prediction logic by computing an average of the predicted
  outcomes of the sub-estimators.

- The ``RandomForestClassifier`` and ``RandomForestRegressor`` derived
  classes provide the user with concrete implementations of
  the forest ensemble method using classical, deterministic
  ``DecisionTreeClassifier`` and ``DecisionTreeRegressor`` as
  sub-estimator implementations.

- The ``ExtraTreesClassifier`` and ``ExtraTreesRegressor`` derived
  classes provide the user with concrete implementations of the
  forest ensemble method using the extremely randomized trees
  ``ExtraTreeClassifier`` and ``ExtraTreeRegressor`` as
  sub-estimator implementations.

Single and multi-output problems are both handled.
"""

# TODO: Documentation.

# Author: Pedro Ilidio <pedrilidio@gmail.com>
# Adapted from scikit-learn.
#
# License: BSD 3 clause

from typing import Iterable
from warnings import warn

from abc import ABCMeta
import numpy as np
from scipy.sparse import issparse

from sklearn.base import is_classifier, _fit_context
from sklearn.tree._tree import DTYPE, DOUBLE
from sklearn.utils import check_random_state, compute_sample_weight
from sklearn.exceptions import DataConversionWarning
from sklearn.utils.parallel import Parallel, delayed
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import _check_sample_weight
from sklearn.ensemble._forest import _get_n_samples_bootstrap
from sklearn.ensemble._forest import _generate_sample_indices
from sklearn.ensemble._forest import BaseForest, ForestRegressor

from ..melter import row_cartesian_product
from ..tree import (
    BipartiteDecisionTreeRegressor,
    BipartiteExtraTreeRegressor,
)
from ..base import BaseBipartiteEstimator
from ..utils import _X_is_multipartite

__all__ = [
    "BipartiteRandomForestRegressor",
    "BipartiteExtraTreesRegressor",
]

MAX_INT = np.iinfo(np.int32).max


def _get_n_samples_bootstrap_nd(n_samples, max_samples):
    """
    Get the number of samples in a bootstrap sample for each axis.

    Parameters
    ----------
    n_samples : list-like[int]
        Number of samples in each axis of the dataset. Usually y.shape
    max_samples : int or float or list-like[int or float]
        For each axis, the maximum number of samples to draw from the total
        available. If a single value is given, that same value is
        considered for all the axes.
            - if float, this indicates a fraction of the total and should be
              the interval `(0.0, 1.0]`;
            - if int, this indicates the exact number of samples;
            - if None, this indicates the total number of samples.

    Returns
    -------
    n_samples_bootstrap : tuple[int]
        The total number of samples to draw from each axis for the
        bootstrap sample.
    """
    # NOTE: This strategy is inherently different from the monopartite version,
    # being naturally more agressive.
    if isinstance(max_samples, float):  # TODO: remove?
        max_samples = max_samples ** (1/len(n_samples))

    if not isinstance(max_samples, Iterable):
        max_samples = [max_samples for _ in n_samples]

    return tuple(
        _get_n_samples_bootstrap(ns, ms)
        for ns, ms in zip(n_samples, max_samples)
    )


def _generate_sample_indices_nd(random_state, n_samples, n_samples_bootstrap):
    sample_indices = []
    # Sample indices values for all axes are considered sequentially in a single
    # array. In 2D, for instance, row (axis 0) samples have indices from 0 to
    # n_samples[0]-1 and column (axis 1) samples are those with the indices from
    # n_samples[0] to n_samples[0]+n_samples[1].
    steps = [0] + list(np.cumsum(n_samples)[:-1])
    for ns, nsb, step in zip(n_samples, n_samples_bootstrap, steps):
        sample_indices.append(
            _generate_sample_indices(random_state, ns, nsb) + step
        )
    return np.hstack(sample_indices)


def _generate_unsampled_indices_nd(random_state, n_samples, n_samples_bootstrap):
    """
    Private function used to forest._set_oob_score function."""
    total_n_samples = np.sum(n_samples)

    sample_indices = _generate_sample_indices_nd(
        random_state, n_samples, n_samples_bootstrap
    )
    sample_counts = np.bincount(sample_indices, minlength=total_n_samples)
    unsampled_mask = sample_counts == 0
    indices_range = np.arange(total_n_samples)
    unsampled_indices = indices_range[unsampled_mask]

    # Split the indices for each axis.
    # FIXME: no better way?
    steps = [0] + list(np.cumsum(n_samples))
    ax_unsampled_indices = []
    for i in range(len(steps)-1):
        mask = (steps[i] < unsampled_indices) & (unsampled_indices < steps[i+1])
        ui = unsampled_indices[mask] - steps[i]
        ax_unsampled_indices.append(ui)

    return ax_unsampled_indices


def _parallel_build_trees_nd(
    tree,
    bootstrap,
    X,
    y,
    sample_weight,
    tree_idx,
    n_trees,
    verbose=0,
    class_weight=None,
    n_samples_bootstrap=None,
    tree_fit_params=None,
):
    """
    Private function used to fit a single tree in parallel."""
    if verbose > 1:
        print("building tree %d of %d" % (tree_idx + 1, n_trees))
    
    tree_fit_params = tree_fit_params or {}

    if bootstrap:
        n_samples = y.shape  # FIXME: not valid for multi-output.
        total_n_samples = np.sum(n_samples)

        if sample_weight is None:
            curr_sample_weight = np.ones((total_n_samples,), dtype=np.float64)
        else:
            curr_sample_weight = sample_weight.copy()

        indices = _generate_sample_indices_nd(
            tree.random_state, n_samples, n_samples_bootstrap
        )
        sample_counts = np.bincount(indices, minlength=total_n_samples)
        curr_sample_weight *= sample_counts

        if class_weight == "balanced_subsample":
            curr_sample_weight *= compute_sample_weight(
                "balanced",
                y.reshape(-1),
                indices=indices,
            )

        tree.fit(
            X, y,
            sample_weight=curr_sample_weight,
            check_input=False,
            **tree_fit_params,
        )
    else:
        tree.fit(
            X, y,
            sample_weight=sample_weight,
            check_input=False,
            **tree_fit_params,
        )

    return tree


class BaseMultipartiteForest(
    BaseForest,
    BaseBipartiteEstimator,
    metaclass=ABCMeta,
):
    """
    Base class for forests of ND trees.

    Warning: This class should not be used directly. Use derived classes
    instead.
    """

    _parameter_constraints: dict = {
        **BaseForest._parameter_constraints,
    }

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y, sample_weight=None, tree_fit_params=None):
        """
        Build a forest of trees from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, its dtype will be converted
            to ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in
            regression).

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. In the case of
            classification, splits are also ignored if they would result in any
            single class carrying a negative weight in either child node.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        # Validate or convert input data
        if issparse(y):
            raise ValueError("sparse multilabel-indicator for y is not supported.")
        X, y = self._validate_data(
            X, y, accept_sparse="csc", dtype=DTYPE,  # multi_output=True, 
        )

        # FIXME: is validation correct?
        ### Copied from ..tree._classes.BaseDecisionTree 
        # # Need to validate separately here.
        # # We can't pass multi_output=True because that would allow y to be
        # # csr.
        # check_X_params = dict(dtype=DTYPE, accept_sparse="csc")
        # check_y_params = dict(dtype=None, ensure_2d=False)
        # X, y = self._validate_data(
        #     X, y, validate_separately=(check_X_params, check_y_params)
        # )

        if sample_weight is not None:
            sample_weight = _check_sample_weight(
                sample_weight, row_cartesian_product(X),
            )

        for ax in range(len(X)):
            if issparse(X[ax]):
                # Pre-sort indices to avoid that each individual tree of the
                # ensemble sorts the indices.
                X[ax].sort_indices()

        y = np.atleast_1d(y)
        if y.ndim == 2 and y.shape[1] == 1:
            warn(
                "A column-vector y was passed when a 1d array was"
                " expected. Please change the shape of y to "
                "(n_samples,), for example using ravel().",
                DataConversionWarning,
                stacklevel=2,
            )

        if y.ndim == 1:
            # reshape is necessary to preserve the data contiguity against vs
            # [:, np.newaxis] that does not.
            y = np.reshape(y, (-1, 1))

        if self.criterion == "poisson":
            if np.any(y < 0):
                raise ValueError(
                    "Some value(s) of y are negative which is "
                    "not allowed for Poisson regression."
                )
            if np.sum(y) <= 0:
                raise ValueError(
                    "Sum of y is not strictly positive which "
                    "is necessary for Poisson regression."
                )

        # self.n_outputs_ = y.shape[1]
        if self.prediction_weights == "raw":
            self.n_outputs_ = np.sum(y.shape)  # TODO: bipartite multioutput
        else:
            self.n_outputs_ = 1

        y, expanded_class_weight = self._validate_y_class_weight(y)

        if getattr(y, "dtype", None) != DOUBLE or not y.flags.contiguous:
            y = np.ascontiguousarray(y, dtype=DOUBLE)

        if expanded_class_weight is not None:
            if sample_weight is not None:
                sample_weight = sample_weight * expanded_class_weight
            else:
                sample_weight = expanded_class_weight

        if not self.bootstrap and self.max_samples is not None:
            raise ValueError(
                "`max_sample` cannot be set if `bootstrap=False`. "
                "Either switch to `bootstrap=True` or set "
                "`max_sample=None`."
            )
        elif self.bootstrap:
            n_samples_bootstrap = _get_n_samples_bootstrap_nd(
                n_samples=y.shape,  # FIXME: will change for multi-output. 
                max_samples=self.max_samples
            )
        else:
            n_samples_bootstrap = None

        # Check parameters
        self._validate_estimator()

        if not self.bootstrap and self.oob_score:
            raise ValueError("Out of bag estimation only available if bootstrap=True")

        random_state = check_random_state(self.random_state)

        if not self.warm_start or not hasattr(self, "estimators_"):
            # Free allocated memory, if any
            self.estimators_ = []

        n_more_estimators = self.n_estimators - len(self.estimators_)

        if n_more_estimators < 0:
            raise ValueError(
                "n_estimators=%d must be larger or equal to "
                "len(estimators_)=%d when warm_start==True"
                % (self.n_estimators, len(self.estimators_))
            )

        elif n_more_estimators == 0:
            warn(
                "Warm-start fitting without increasing n_estimators does not "
                "fit new trees."
            )
        else:
            if self.warm_start and len(self.estimators_) > 0:
                # We draw from the random state to get the random state we
                # would have got if we hadn't used a warm_start.
                random_state.randint(MAX_INT, size=len(self.estimators_))

            trees = [
                self._make_estimator(append=False, random_state=random_state)
                for i in range(n_more_estimators)
            ]

            # Parallel loop: we prefer the threading backend as the Cython code
            # for fitting the trees is internally releasing the Python GIL
            # making threading more efficient than multiprocessing in
            # that case. However, for joblib 0.12+ we respect any
            # parallel_backend contexts set at a higher level,
            # since correctness does not rely on using threads.
            trees = Parallel(
                n_jobs=self.n_jobs,
                verbose=self.verbose,
                prefer="threads",
            )(
                delayed(_parallel_build_trees_nd)(
                    t,
                    self.bootstrap,
                    X,
                    y,
                    sample_weight,
                    i,
                    len(trees),
                    verbose=self.verbose,
                    class_weight=self.class_weight,
                    n_samples_bootstrap=n_samples_bootstrap,
                    tree_fit_params=tree_fit_params,
                )
                for i, t in enumerate(trees)
            )

            # Collect newly grown trees
            self.estimators_.extend(trees)

        if self.oob_score:
            y_ = y.reshape(-1)
            X_ = row_cartesian_product(X)

            y_type = type_of_target(y_)
            if y_type in ("multiclass-multioutput", "unknown"):
                # FIXME: we could consider to support multiclass-multioutput if
                # we introduce or reuse a constructor parameter (e.g.
                # oob_score) allowing our user to pass a callable defining the
                # scoring strategy on OOB sample. (sklearn)
                raise ValueError(
                    "The type of target cannot be used to compute OOB "
                    f"estimates. Got {y_type} while only the following are "
                    "supported: continuous, continuous-multioutput, binary, "
                    "multiclass, multilabel-indicator."
                )
            self._set_oob_score_and_attributes(X_, y_)

        # Decapsulate classes_ attributes
        if hasattr(self, "classes_") and self.n_outputs_ == 1:
            self.n_classes_ = self.n_classes_[0]
            self.classes_ = self.classes_[0]

        return self

    # TODO: avoid boilerplate from sklearn
    def _compute_oob_predictions(self, X, y):
        """Compute and set the OOB score.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix.
        y : ndarray of shape (n_samples, n_outputs)
            The target matrix.

        Returns
        -------
        oob_pred : ndarray of shape (n_samples, n_classes, n_outputs) or \
                (n_samples, 1, n_outputs)
            The OOB predictions.
        """
        # Prediction requires X to be in CSR format
        X = [Xi.tocsr() if issparse(Xi) else Xi for Xi in X]

        n_samples = y.shape
        n_outputs = self.n_outputs_

        if is_classifier(self) and hasattr(self, "n_classes_"):
            # n_classes_ is a ndarray at this stage
            # all the supported type of target will have the same number of
            # classes in all outputs
            oob_pred_shape = (n_samples, self.n_classes_[0], n_outputs)
        else:
            # for regression, n_classes_ does not exist and we create an empty
            # axis to be consistent with the classification case and make
            # the array operations compatible with the 2 settings
            oob_pred_shape = (n_samples, 1, n_outputs)

        oob_pred = np.zeros(shape=oob_pred_shape, dtype=np.float64)
        n_oob_pred = np.zeros((n_samples, n_outputs), dtype=np.int64)

        n_samples_bootstrap = _get_n_samples_bootstrap_nd(
            n_samples,
            self.max_samples,
        )
        for estimator in self.estimators_:
            # Separated for each axis.
            unsampled_indices = _generate_unsampled_indices_nd(
                estimator.random_state,
                n_samples,
                n_samples_bootstrap,
            )

            X_unsampled = [Xi[ui] for Xi, ui in zip(X, unsampled_indices)]
            y_pred = self._get_oob_predictions(estimator, X_unsampled)
            oob_pred[unsampled_indices, ...] += y_pred
            n_oob_pred[unsampled_indices, :] += 1

        for k in range(n_outputs):
            if (n_oob_pred == 0).any():
                warn(
                    "Some inputs do not have OOB scores. This probably means "
                    "too few trees were used to compute any reliable OOB "
                    "estimates.",
                    UserWarning,
                )
                n_oob_pred[n_oob_pred == 0] = 1
            oob_pred[..., k] /= n_oob_pred[..., [k]]

        return oob_pred

    def _validate_X_predict(self, X):
        """ Validate X whenever one tries to predict, apply, predict_proba.
        """
        if _X_is_multipartite(X):
            X = row_cartesian_product(X)

        return super()._validate_X_predict(X)


class BipartiteRandomForestRegressor(
    BaseMultipartiteForest,
    ForestRegressor,
):
    """
    A random forest regressor.

    A random forest is a meta estimator that fits a number of classifying
    decision trees on various sub-samples of the dataset and uses averaging
    to improve the predictive accuracy and control over-fitting.
    The sub-sample size is controlled with the `max_samples` parameter if
    `bootstrap=True` (default), otherwise the whole dataset is used to build
    each tree.

    Read more in the :ref:`User Guide <forest>`.

    Parameters
    ----------
    n_estimators : int, default=100
        The number of trees in the forest.

        .. versionchanged:: 0.22
           The default value of ``n_estimators`` changed from 10 to 100
           in 0.22.

    criterion : {"squared_error", "absolute_error", "poisson"}, \
            default="squared_error"
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to
        variance reduction as feature selection criterion, "absolute_error"
        for the mean absolute error, and "poisson" which uses reduction in
        Poisson deviance to find splits.
        Training using "absolute_error" is significantly slower
        than when using "squared_error".

        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.

        .. versionadded:: 1.0
           Poisson criterion.

        .. deprecated:: 1.0
            Criterion "mse" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="squared_error"` which is equivalent.

        .. deprecated:: 1.0
            Criterion "mae" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="absolute_error"` which is equivalent.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : {"sqrt", "log2", None}, int or float, default=None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `round(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None or 1.0, then `max_features=n_features`.

        .. note::
            The default of 1.0 is equivalent to bagged trees and more
            randomness can be achieved by setting smaller values, e.g. 0.3.

        .. versionchanged:: 1.1
            The default of `max_features` changed from `"auto"` to 1.0.

        .. deprecated:: 1.1
            The `"auto"` option was deprecated in 1.1 and will be removed
            in 1.3.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int, default=None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    bootstrap : bool, default=True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    oob_score : bool, default=False
        Whether to use out-of-bag samples to estimate the generalization score.
        Only available if bootstrap=True.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors. See :term:`Glossary
        <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).
        See :term:`Glossary <random_state>` for details.

    verbose : int, default=0
        Controls the verbosity when fitting and predicting.

    warm_start : bool, default=False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`the Glossary <warm_start>`.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

        .. versionadded:: 0.22

    max_samples : int or float, default=None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

        .. versionadded:: 0.22

    Attributes
    ----------
    estimator_ : DecisionTreeRegressor
        The child estimator template used to create the collection of fitted
        sub-estimators.

    estimators_ : list of DecisionTreeRegressor
        The collection of fitted sub-estimators.

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    n_features_ : int
        The number of features when ``fit`` is performed.

        .. deprecated:: 1.0
            Attribute `n_features_` was deprecated in version 1.0 and will be
            removed in 1.2. Use `n_features_in_` instead.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

        .. versionadded:: 1.0

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
        This attribute exists only when ``oob_score`` is True.

    oob_prediction_ : ndarray of shape (n_samples,) or (n_samples, n_outputs)
        Prediction computed with out-of-bag estimate on the training set.
        This attribute exists only when ``oob_score`` is True.

    See Also
    --------
    sklearn.tree.DecisionTreeRegressor : A decision tree regressor.
    sklearn.ensemble.ExtraTreesRegressor : Ensemble of extremely randomized
        tree regressors.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    The features are always randomly permuted at each split. Therefore,
    the best found split may vary, even with the same training data,
    ``max_features=n_features`` and ``bootstrap=False``, if the improvement
    of the criterion is identical for several splits enumerated during the
    search of the best split. To obtain a deterministic behaviour during
    fitting, ``random_state`` has to be fixed.

    The default value ``max_features="auto"`` uses ``n_features``
    rather than ``n_features / 3``. The latter was originally suggested in
    [1], whereas the former was more recently justified empirically in [2].

    References
    ----------
    .. [1] L. Breiman, "Random Forests", Machine Learning, 45(1), 5-32, 2001.

    .. [2] P. Geurts, D. Ernst., and L. Wehenkel, "Extremely randomized
           trees", Machine Learning, 63(1), 3-42, 2006.

    Examples
    --------
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from sklearn.datasets import make_regression
    >>> X, y = make_regression(n_features=4, n_informative=2,
    ...                        random_state=0, shuffle=False)
    >>> regr = RandomForestRegressor(max_depth=2, random_state=0)
    >>> regr.fit(X, y)
    RandomForestRegressor(...)
    >>> print(regr.predict([[0, 0, 0, 0]]))
    [-8.32987858]
    """

    _parameter_constraints: dict = {
        **ForestRegressor._parameter_constraints,
        **BipartiteDecisionTreeRegressor._parameter_constraints,
    }
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
        prediction_weights=None,
    ):
        super().__init__(
            estimator=BipartiteDecisionTreeRegressor(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Bipartite parameters:
                "min_rows_split",
                "min_cols_split",
                "min_rows_leaf",
                "min_cols_leaf",
                "min_row_weight_fraction_leaf",
                "min_col_weight_fraction_leaf",
                "max_row_features",
                "max_col_features",
                "bipartite_adapter",
                "prediction_weights",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        self.min_rows_split = min_rows_split
        self.min_cols_split = min_cols_split
        self.min_rows_leaf = min_rows_leaf
        self.min_cols_leaf = min_cols_leaf
        self.min_row_weight_fraction_leaf = min_row_weight_fraction_leaf
        self.min_col_weight_fraction_leaf = min_col_weight_fraction_leaf
        self.max_row_features = max_row_features
        self.max_col_features = max_col_features
        self.bipartite_adapter = bipartite_adapter
        self.prediction_weights = prediction_weights


class BipartiteExtraTreesRegressor(
    BaseMultipartiteForest,
    ForestRegressor,
):
    """
    An extra-trees regressor.

    This class implements a meta estimator that fits a number of
    randomized decision trees (a.k.a. extra-trees) on various sub-samples
    of the dataset and uses averaging to improve the predictive accuracy
    and control over-fitting.

    Read more in the :ref:`User Guide <forest>`.

    Parameters
    ----------
    n_estimators : int, default=100
        The number of trees in the forest.

        .. versionchanged:: 0.22
           The default value of ``n_estimators`` changed from 10 to 100
           in 0.22.

    criterion : {"squared_error", "absolute_error"}, default="squared_error"
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to
        variance reduction as feature selection criterion, and "absolute_error"
        for the mean absolute error.

        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.

        .. deprecated:: 1.0
            Criterion "mse" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="squared_error"` which is equivalent.

        .. deprecated:: 1.0
            Criterion "mae" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="absolute_error"` which is equivalent.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : {"sqrt", "log2", None}, int or float, default=None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `round(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None or 1.0, then `max_features=n_features`.

        .. note::
            The default of 1.0 is equivalent to bagged trees and more
            randomness can be achieved by setting smaller values, e.g. 0.3.

        .. versionchanged:: 1.1
            The default of `max_features` changed from `"auto"` to 1.0.

        .. deprecated:: 1.1
            The `"auto"` option was deprecated in 1.1 and will be removed
            in 1.3.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int, default=None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    bootstrap : bool, default=False
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    oob_score : bool, default=False
        Whether to use out-of-bag samples to estimate the generalization score.
        Only available if bootstrap=True.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors. See :term:`Glossary
        <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls 3 sources of randomness:

        - the bootstrapping of the samples used when building trees
          (if ``bootstrap=True``)
        - the sampling of the features to consider when looking for the best
          split at each node (if ``max_features < n_features``)
        - the draw of the splits for each of the `max_features`

        See :term:`Glossary <random_state>` for details.

    verbose : int, default=0
        Controls the verbosity when fitting and predicting.

    warm_start : bool, default=False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`the Glossary <warm_start>`.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

        .. versionadded:: 0.22

    max_samples : int or float, default=None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

        .. versionadded:: 0.22

    Attributes
    ----------
    estimator_ : ExtraTreeRegressor
        The child estimator template used to create the collection of fitted
        sub-estimators.

    estimators_ : list of DecisionTreeRegressor
        The collection of fitted sub-estimators.

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    n_features_ : int
        The number of features.

        .. deprecated:: 1.0
            Attribute `n_features_` was deprecated in version 1.0 and will be
            removed in 1.2. Use `n_features_in_` instead.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

        .. versionadded:: 1.0

    n_outputs_ : int
        The number of outputs.

    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
        This attribute exists only when ``oob_score`` is True.

    oob_prediction_ : ndarray of shape (n_samples,) or (n_samples, n_outputs)
        Prediction computed with out-of-bag estimate on the training set.
        This attribute exists only when ``oob_score`` is True.

    See Also
    --------
    ExtraTreesClassifier : An extra-trees classifier with random splits.
    RandomForestClassifier : A random forest classifier with optimal splits.
    RandomForestRegressor : Ensemble regressor using trees with optimal splits.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    References
    ----------
    .. [1] P. Geurts, D. Ernst., and L. Wehenkel, "Extremely randomized trees",
           Machine Learning, 63(1), 3-42, 2006.

    Examples
    --------
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.ensemble import ExtraTreesRegressor
    >>> X, y = load_diabetes(return_X_y=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...     X, y, random_state=0)
    >>> reg = ExtraTreesRegressor(n_estimators=100, random_state=0).fit(
    ...    X_train, y_train)
    >>> reg.score(X_test, y_test)
    0.2727...
    """

    _parameter_constraints: dict = {
        **ForestRegressor._parameter_constraints,
        **BipartiteDecisionTreeRegressor._parameter_constraints,
    }
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
        prediction_weights=None,
    ):
        super().__init__(
            estimator=BipartiteExtraTreeRegressor(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Bipartite parameters:
                "min_rows_split",
                "min_cols_split",
                "min_rows_leaf",
                "min_cols_leaf",
                "min_row_weight_fraction_leaf",
                "min_col_weight_fraction_leaf",
                "max_row_features",
                "max_col_features",
                "bipartite_adapter",
                "prediction_weights",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        self.min_rows_split = min_rows_split
        self.min_cols_split = min_cols_split
        self.min_rows_leaf = min_rows_leaf
        self.min_cols_leaf = min_cols_leaf
        self.min_row_weight_fraction_leaf = min_row_weight_fraction_leaf
        self.min_col_weight_fraction_leaf = min_col_weight_fraction_leaf
        self.max_row_features = max_row_features
        self.max_col_features = max_col_features
        self.bipartite_adapter = bipartite_adapter
        self.prediction_weights = prediction_weights



# TODO: Documentation.
# TODO: Classifiers.
from abc import ABCMeta
from sklearn.ensemble._forest import (
    BaseForest,
    ForestRegressor,
    ForestClassifier,
)
from sklearn.utils._param_validation import StrOptions
from ..tree._semisupervised_classes import (
    ExtraTreeClassifierSS,
    DecisionTreeClassifierSS,
    DecisionTreeRegressorSS,
    ExtraTreeRegressorSS,
    BipartiteDecisionTreeRegressorSS,
    BipartiteExtraTreeRegressorSS,
)
from ._forest import BaseMultipartiteForest

__all__ = [
    "ExtraTreesRegressorSS",
    "RandomForestRegressorSS",
    "BipartiteRandomForestRegressorSS",
    "BipartiteExtraTreesRegressorSS",
]


class BaseSemisupervisedForest(
    BaseForest,
    metaclass=ABCMeta,
):
    def fit(self, X, y, sample_weight=None, _X_targets=None):
        # TODO: Since we cannot pass _X_targets to trees in an ensemble,
        # inside sklearn.ensemble._forest._parallel_build_trees(),
        # we add _X_targets as a parameter in the forest.estimator_params
        # list, that will be reused by the individual trees without copying
        # the array for each tree and consuming a large amount of memory.
        # In forests, check_input is set to false, in which case we do not
        # override self._X_targets below.
        self._X_targets = self._check_X_targets(X, _X_targets)

        return super().fit(X, y, sample_weight)

    def _check_X_targets(self, X, _X_targets=None):
        self._validate_estimator()
        self._X_targets = None
        tree = self._make_estimator(append=False)
        return tree._check_X_targets(X, _X_targets)


class RandomForestClassifierSS(
    BaseSemisupervisedForest,
    ForestClassifier,
):

    _parameter_constraints: dict = {
        **ForestClassifier._parameter_constraints,
        **DecisionTreeClassifierSS._parameter_constraints,
        "class_weight": [
            StrOptions({"balanced_subsample", "balanced"}),
            dict,
            list,
            None,
        ],
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=None,
        # Semi-supervised parameters:
        unsupervised_criterion="squared_error",
        supervision=0.5,
        update_supervision=None,
        ss_adapter=None,
        preprocess_X_targets=None,
    ):
        super().__init__(
            estimator=DecisionTreeClassifierSS(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Semi-supervised parameters:
                "_X_targets",
                "supervision",
                "unsupervised_criterion",
                "update_supervision",
                "ss_adapter",
                "preprocess_X_targets",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        # Semi-supervised parameters:
        self.unsupervised_criterion = unsupervised_criterion
        self.supervision = supervision
        self.update_supervision = update_supervision
        self.ss_adapter = ss_adapter
        self.preprocess_X_targets = preprocess_X_targets


class ExtraTreesClassifierSS(
    BaseSemisupervisedForest,
    ForestClassifier,
):
    _parameter_constraints: dict = {
        **ForestClassifier._parameter_constraints,
        **ExtraTreeClassifierSS._parameter_constraints,
        "class_weight": [
            StrOptions({"balanced_subsample", "balanced"}),
            dict,
            list,
            None,
        ],
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=None,
        # Semi-supervised parameters:
        unsupervised_criterion="squared_error",
        supervision=0.5,
        update_supervision=None,
        ss_adapter=None,
        preprocess_X_targets=None,
    ):
        super().__init__(
            estimator=ExtraTreeClassifierSS(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Semi-supervised parameters:
                "supervision",
                "unsupervised_criterion",
                "update_supervision",
                "ss_adapter",
                "preprocess_X_targets",
                "_X_targets",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        # Semi-supervised parameters:
        self.unsupervised_criterion = unsupervised_criterion
        self.supervision = supervision
        self.update_supervision = update_supervision
        self.ss_adapter = ss_adapter
        self.preprocess_X_targets = preprocess_X_targets


class RandomForestRegressorSS(
    BaseSemisupervisedForest,
    ForestRegressor,
):
    _parameter_constraints: dict = {
        **ForestRegressor._parameter_constraints,
        **DecisionTreeRegressorSS._parameter_constraints,
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        # Semi-supervised parameters:
        unsupervised_criterion="squared_error",
        supervision=0.5,
        update_supervision=None,
        ss_adapter=None,
        preprocess_X_targets=None,
    ):
        super().__init__(
            estimator=DecisionTreeRegressorSS(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Semi-supervised parameters:
                "supervision",
                "unsupervised_criterion",
                "update_supervision",
                "ss_adapter",
                "preprocess_X_targets",
                "_X_targets",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        # Semi-supervised parameters:
        self.unsupervised_criterion = unsupervised_criterion
        self.supervision = supervision
        self.update_supervision = update_supervision
        self.ss_adapter = ss_adapter
        self.preprocess_X_targets = preprocess_X_targets


class ExtraTreesRegressorSS(
    BaseSemisupervisedForest,
    ForestRegressor,
):
    _parameter_constraints: dict = {
        **ForestRegressor._parameter_constraints,
        **ExtraTreeRegressorSS._parameter_constraints,
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        # Semi-supervised parameters:
        unsupervised_criterion="squared_error",
        supervision=0.5,
        update_supervision=None,
        ss_adapter=None,
        preprocess_X_targets=None,
    ):
        super().__init__(
            estimator=ExtraTreeRegressorSS(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Semi-supervised parameters:
                "supervision",
                "unsupervised_criterion",
                "update_supervision",
                "ss_adapter",
                "preprocess_X_targets",
                "_X_targets",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        # Semi-supervised parameters:
        self.unsupervised_criterion = unsupervised_criterion
        self.supervision = supervision
        self.update_supervision = update_supervision
        self.ss_adapter = ss_adapter
        self.preprocess_X_targets = preprocess_X_targets


class BaseSemisupervisedBipartiteForest(
    BaseMultipartiteForest,
    metaclass=ABCMeta,
):
    def fit(self, X, y, sample_weight=None, check_input=True, _X_targets=None):
        if check_input:
            # _X_targets will be reused by the individual trees without copying
            # the array in each of them and consuming a large amount of memory.
            _X_targets = self._check_X_targets(X, _X_targets)

        self._X_targets = _X_targets  # TODO: This is not necessary. Keep it?

        return super().fit(
            X, y,
            sample_weight=sample_weight,
            tree_fit_params={'_X_targets': _X_targets},
        )

    def _check_X_targets(self, X, _X_targets=None):
        self._validate_estimator()
        tree = self._make_estimator(append=False)
        return tree._check_X_targets(X, _X_targets)

    def _more_tags(self):
        self._validate_estimator()
        tree = self._make_estimator(append=False)
        return tree._more_tags()


class BipartiteRandomForestRegressorSS(
    BaseSemisupervisedBipartiteForest,
    ForestRegressor,
):

    _parameter_constraints: dict = {
        **BaseSemisupervisedBipartiteForest._parameter_constraints,
        **BipartiteDecisionTreeRegressorSS._parameter_constraints,
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        update_supervision=None,
        preprocess_X_targets=None,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
        prediction_weights=None,
        # Biaprtite and semisupervised parameters:
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        axis_decision_only=False,
    ):
        super().__init__(
            estimator=BipartiteDecisionTreeRegressorSS(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Semi-supervised parameters:
                "supervision",
                "unsupervised_criterion_rows",
                "unsupervised_criterion_cols",
                "update_supervision",
                "ss_adapter",
                "axis_decision_only",
                "preprocess_X_targets",
                # Bipartite parameters:
                "min_rows_split",
                "min_cols_split",
                "min_rows_leaf",
                "min_cols_leaf",
                "min_row_weight_fraction_leaf",
                "min_col_weight_fraction_leaf",
                "max_row_features",
                "max_col_features",
                "bipartite_adapter",
                "prediction_weights",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        # Semi-supervised parameters:
        self.supervision = supervision
        self.ss_adapter = ss_adapter
        self.unsupervised_criterion_rows = unsupervised_criterion_rows
        self.unsupervised_criterion_cols = unsupervised_criterion_cols
        self.update_supervision = update_supervision
        self.axis_decision_only = axis_decision_only
        self.preprocess_X_targets = preprocess_X_targets

        # Bipartite parameters:
        self.min_rows_split = min_rows_split
        self.min_cols_split = min_cols_split
        self.min_rows_leaf = min_rows_leaf
        self.min_cols_leaf = min_cols_leaf
        self.min_row_weight_fraction_leaf = min_row_weight_fraction_leaf
        self.min_col_weight_fraction_leaf = min_col_weight_fraction_leaf
        self.max_row_features = max_row_features
        self.max_col_features = max_col_features
        self.bipartite_adapter = bipartite_adapter
        self.prediction_weights = prediction_weights


class BipartiteExtraTreesRegressorSS(
    BaseSemisupervisedBipartiteForest,
    ForestRegressor,
):

    _parameter_constraints: dict = {
        **BaseSemisupervisedBipartiteForest._parameter_constraints,
        **BipartiteExtraTreeRegressorSS._parameter_constraints,
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=False,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        update_supervision=None,
        axis_decision_only=False,
        prediction_weights=None,
        preprocess_X_targets=None,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
    ):
        super().__init__(
            estimator=BipartiteExtraTreeRegressorSS(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                # Semi-supervised parameters:
                "supervision",
                "unsupervised_criterion_rows",
                "unsupervised_criterion_cols",
                "update_supervision",
                "ss_adapter",
                "axis_decision_only",
                "preprocess_X_targets",
                # Bipartite parameters:
                "min_rows_split",
                "min_cols_split",
                "min_rows_leaf",
                "min_cols_leaf",
                "min_row_weight_fraction_leaf",
                "min_col_weight_fraction_leaf",
                "max_row_features",
                "max_col_features",
                "bipartite_adapter",
                "prediction_weights",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        # Semi-supervised parameters:
        self.supervision = supervision
        self.ss_adapter = ss_adapter
        self.unsupervised_criterion_rows = unsupervised_criterion_rows
        self.unsupervised_criterion_cols = unsupervised_criterion_cols
        self.update_supervision = update_supervision
        self.axis_decision_only = axis_decision_only
        self.preprocess_X_targets = preprocess_X_targets

        # Bipartite parameters:
        self.min_rows_split = min_rows_split
        self.min_cols_split = min_cols_split
        self.min_rows_leaf = min_rows_leaf
        self.min_cols_leaf = min_cols_leaf
        self.min_row_weight_fraction_leaf = min_row_weight_fraction_leaf
        self.min_col_weight_fraction_leaf = min_col_weight_fraction_leaf
        self.max_row_features = max_row_features
        self.max_col_features = max_col_features
        self.bipartite_adapter = bipartite_adapter
        self.prediction_weights = prediction_weights



import pytest
from pprint import pprint
from typing import Sequence
import numpy as np
from sklearn.utils._tags import _safe_tags
from sklearn.model_selection import KFold
from sklearn.metrics.pairwise import rbf_kernel

from bipartite_learn.model_selection import (
    multipartite_cross_validate,
    check_multipartite_cv,
    MultipartiteCrossValidator,
)
from bipartite_learn.tree import BipartiteExtraTreeRegressor
from bipartite_learn.ensemble import BipartiteExtraTreesRegressor

import bipartite_learn.ensemble

from .utils.test_utils import stopwatch, parse_args
from .utils.make_examples import make_interaction_regression


CV_DEF_PARAMS = dict(
    groups=None,
    scoring="average_precision",
    cv=3,
    n_jobs=None,
    verbose=10,
    fit_params=None,
    pre_dispatch="2*n_jobs",
    return_train_score=False,
    error_score="raise",
    diagonal=False,
    train_test_combinations=None,
    pairwise=False,
)


@pytest.fixture(params=range(5))
def random_state(request):
    return request.param


@pytest.fixture
def data(random_state):
    XX, Y = make_interaction_regression(
        n_samples=(50, 40),
        n_features=(10, 9),
        return_molten=False,
        noise=0.0,
    )
    XX = [rbf_kernel(Xi, gamma=1) for Xi in XX]
    Y = (Y > Y.mean()).astype("float64")
    return XX, Y


def _test_cv(estimator, data, cv_params=None):
    X, y = data

    cv_params = CV_DEF_PARAMS | (cv_params or {})
    cv_params |= dict(return_estimator=True)

    is_pairwise = cv_params["pairwise"] or _safe_tags(estimator, "pairwise")
    cv = cv_params["cv"]

    if isinstance(cv, Sequence):
        if not isinstance(cv[0], int):
            raise TypeError("cv_params['cv'] must be (sequence of) int")

    elif not isinstance(cv, int):
        raise TypeError("cv_params['cv'] must be (sequence of) int")

    else:  # isinstance(cv_params["cv"], int)
        cv = (cv, cv)

    cv_res = multipartite_cross_validate(
        estimator,
        X=X,
        y=y,
        **cv_params,
    )
    print("Cross-validation results:")
    pprint(cv_res)

    for estimator in cv_res["estimator"]:
        if is_pairwise:
            # For pairwise data, each fold must be square
            # Sum Xi.n_samples where Xi is a training fold of X
            fold_n_samples = sum(s - (s // cvi) for s, cvi in zip(y.shape, cv))
            # When cv does not divide n_samples perfectly, it distributes the
            # remainders across folds (see np.array_split), so that each fold
            # size can vary on one unit. Naturally, this happens for the n_dim
            # X matrices, so n_features_in_ can be up to n_dim smaller.
            assert (estimator.n_features_in_ - fold_n_samples) <= y.ndim
        else:
            # Sum Yi.n_samples where Yi is a training fold of Y
            n_features = sum(Xi.shape[1] for Xi in X)
            assert estimator.n_features_in_ == n_features

    return cv_res


# parametrize estimator:
@pytest.mark.parametrize(
    "estimator",
    [
        BipartiteExtraTreeRegressor(
            min_samples_leaf=25,
        ),
        BipartiteExtraTreesRegressor(
            min_samples_leaf=25,
        ),
        bipartite_learn.ensemble.BipartiteExtraTreesRegressorSS(
            n_estimators=10,
            min_samples_leaf=25,
            bipartite_adapter="gmosa",
            criterion="squared_error",
        ),
    ],
)
def test_cv_on_estimators(estimator, data, random_state, **PARAMS):
    _test_cv(
        estimator=estimator.set_params(random_state=random_state),
        data=data,
    )


def test_cv_pairwise_tag(data, random_state):
    estimator = BipartiteExtraTreeRegressor(
        min_samples_leaf=25,
        random_state=random_state,
    )

    # Monkey-patch pairwise tag
    old_tags = _safe_tags(estimator)
    type(estimator)._more_tags = lambda self: old_tags | dict(pairwise=True)
    # Revert patch even if an exception is raised
    try:
        _test_cv(
            data=data,
            estimator=estimator,
            cv_params=dict(pairwise=False),  # get tag from the estimator
        )
    finally:
        type(estimator)._more_tags = lambda self: old_tags

    assert not _safe_tags(estimator, "pairwise")


def test_cv_pairwise_parameter(data, random_state):
    estimator = BipartiteExtraTreeRegressor(
        min_samples_leaf=25,
        random_state=random_state,
    )
    _test_cv(
        estimator=estimator,
        data=data,
        cv_params=dict(pairwise=True),
    )

    assert not _safe_tags(estimator, "pairwise")


def test_check_multipartite_cv_with_multipartite_cv():
    # Test that if cv is already a MultipartiteCrossValidator object,
    # the function returns it directly without any modifications
    cv = MultipartiteCrossValidator(KFold(n_splits=2), n_parts=5)
    assert check_multipartite_cv(cv) == cv


def test_check_multipartite_cv_with_n_parts():
    # Test that the function creates a MultipartiteCrossValidator object with
    # n_parts equal to the specified value when cv is an integer
    cv = check_multipartite_cv(3, n_parts=3)
    assert isinstance(cv, MultipartiteCrossValidator)
    assert len(cv.cross_validators) == 3
    assert cv.n_parts == 3


def test_check_multipartite_cv_with_list_of_cv():
    # Test that the function creates a MultipartiteCrossValidator object with
    # the specified cross-validation objects when cv is a list/tuple
    cv1 = KFold(n_splits=2)
    cv2 = KFold(n_splits=3)
    cv = check_multipartite_cv([cv1, cv2])
    assert isinstance(cv, MultipartiteCrossValidator)
    assert len(cv.cross_validators) == 2
    assert cv.cross_validators[0] == cv1
    assert cv.cross_validators[1] == cv2
    assert cv.n_parts == 2


@pytest.mark.parametrize("shuffle", [True, False])
def test_check_multipartite_cv_shuffle(data, shuffle, random_state):
    X, y = data
    n_parts = 2
    n_splits = (3, 4)

    cv = check_multipartite_cv(
        n_splits,
        y=y,
        shuffle=shuffle,
        random_state=random_state if shuffle else None,
        n_parts=n_parts,
    )

    assert isinstance(cv, MultipartiteCrossValidator)
    assert len(cv.cross_validators) == n_parts

    split = list(cv.split(X))

    assert len(split) == np.prod(n_splits)

    for ax_split in split:
        for i in zip(ax_split, y.shape, n_splits):
            (ax_train_idx, ax_test_idx), n_samples, ax_n_splits = i

            is_consecutive = (np.diff(np.array(ax_test_idx)) == 1).all()
            if shuffle:
                assert not is_consecutive
            else:
                assert is_consecutive
            assert (len(ax_test_idx) - n_samples // ax_n_splits) <= 1
            assert len(ax_train_idx) + len(ax_test_idx) == n_samples



from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy

extensions = [
    Extension("tests.utils.tree_utils", ["tests/utils/tree_utils.pyx"],
        extra_compile_args=["-g"],
        extra_link_args=["-g"],
    ),
]

setup(
    name='bipartite_learn_tests',
    ext_modules=cythonize(
        extensions,
        gdb_debug=True,
        annotate=True,
        language_level="3",
    ),
    include_dirs=[numpy.get_include()],
)



import logging
import warnings
import functools

import pytest
from sklearn.utils.validation import check_random_state
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.utils._testing import assert_allclose
from sklearn.metrics import r2_score, f1_score
from sklearn.metrics.pairwise import rbf_kernel
from bipartite_learn.base import BaseMultipartiteEstimator
from bipartite_learn.utils import all_estimators
from bipartite_learn.melter import row_cartesian_product
from .utils.make_examples import (
    make_interaction_blobs,
    make_interaction_regression,
)


def apply_rbf_kernel(regression_data):
    X, Y, x, y = regression_data
    X = [rbf_kernel(X[0]), rbf_kernel(X[1])]
    x = row_cartesian_product(X)
    return X, Y, x, y


@pytest.fixture(params=range(3))
def random_state(request):
    return check_random_state(request.param)


@pytest.fixture
def n_samples(request):
    return (11, 13)


@pytest.fixture
def n_features(request):
    return (5, 7)


@pytest.fixture
def classification_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_blobs(
        return_molten=True,
        n_features=n_features,
        n_samples=n_samples,
        random_state=random_state,
        noise=0.0,
        centers=2,
        row_kwargs={'center_box': (.3, .7), 'cluster_std': .1},
        col_kwargs={'center_box': (.3, .7), 'cluster_std': .1},
    )
    # XXX
    Y = (Y == 0).astype(int)
    y = Y.reshape(-1)
    return X, Y, x, y


@pytest.fixture
def regression_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_targets=None,
        min_target=0.0,
        max_target=1.0,
        noise=0.0,
        return_molten=True,
        return_tree=False,
        random_state=random_state,
        max_depth=None,
    )
    return X, Y, x, y


@pytest.fixture
def pairwise_regression_data(regression_data):
    return apply_rbf_kernel(regression_data)


@pytest.fixture
def pairwise_classification_data(classification_data):
    return apply_rbf_kernel(classification_data)


@pytest.fixture
def regression_scorer():
    return r2_score


@pytest.fixture
def classification_scorer():
    return functools.partial(f1_score, average='macro')


@pytest.fixture
def regression_min_score(regression_data, regression_scorer):
    _, _, x, y = regression_data
    pred = DummyRegressor().fit(x, y).predict(x)
    return regression_scorer(y, pred.reshape(-1, 1))


@pytest.fixture
def classification_min_score(classification_data, classification_scorer):
    _, _, x, y = classification_data
    pred = DummyClassifier().fit(x, y).predict(x)
    return classification_scorer(y, pred.reshape(-1, 1))


@pytest.mark.parametrize(
    'estimator_name, estimator',
    (
        e for e in all_estimators("regressor")
        if not issubclass(e[1], BaseMultipartiteEstimator)
    )
)
def test_all_monopartite_regressors(
    regression_data,
    pairwise_regression_data,
    estimator_name,
    estimator,
    regression_scorer,
    regression_min_score,
):
    estimator = estimator()

    if estimator._get_tags()['pairwise']:
        X, Y, x, y = pairwise_regression_data
    else:
        X, Y, x, y = regression_data

    estimator.fit(X[0], Y)
    estimator.score(X[0], Y)  # Test own scorer
    score = regression_scorer(y, estimator.predict(X[0]).reshape(-1, 1))

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > regression_min_score


@pytest.mark.parametrize(
    'estimator_name, estimator', all_estimators("regressor", "multipartite")
)
def test_all_multipartite_regressors(
    regression_data,
    pairwise_regression_data,
    estimator_name,
    estimator,
    regression_scorer,
    regression_min_score,
):
    estimator = estimator()

    if estimator._get_tags()['pairwise']:
        X, Y, x, y = pairwise_regression_data
    else:
        X, Y, x, y = regression_data

    estimator.fit(X, Y)

    # The two formats must be supported:
    pred1 = estimator.predict(X)
    try:
        pred2 = estimator.predict(x)
        assert_allclose(pred1, pred2)
    except ValueError:
        warnings.warn(f"{estimator_name} does not accept melted input.")

    estimator.score(X, y)  # Test own scorer
    score = regression_scorer(y, estimator.predict(X).reshape(-1, 1))

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > regression_min_score


@pytest.mark.parametrize(
    'estimator_name, estimator', all_estimators("classifier", "multipartite")
)
def test_all_multipartite_classifiers(
    classification_data,
    pairwise_classification_data,
    estimator_name,
    estimator,
    classification_scorer,
    classification_min_score,
):
    estimator = estimator()

    if estimator._get_tags()['pairwise']:
        X, Y, x, y = pairwise_classification_data
    else:
        X, Y, x, y = classification_data

    estimator.fit(X, Y)

    # The two formats must be supported:
    pred1 = estimator.predict(X)
    try:
        pred2 = estimator.predict(x)
        assert_allclose(pred1, pred2)
    except ValueError:
        warnings.warn(f"{estimator_name} does not accept melted input.")

    estimator.score(X, y)  # Test own scorer
    score = classification_scorer(y, estimator.predict(X).reshape(-1, 1))

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > classification_min_score



import logging
from pathlib import Path
import timeit
import numpy as np
from functools import partial
import pytest

from sklearn.tree import DecisionTreeRegressor
from sklearn.base import clone
from bipartite_learn.base import BaseMultipartiteEstimator
from bipartite_learn.tree import BipartiteDecisionTreeRegressor
from bipartite_learn.tree._splitter_factory import (
    make_bipartite_ss_splitter,
    make_semisupervised_criterion,
) 
from bipartite_learn.tree._semisupervised_classes import (
    DecisionTreeRegressorSS, BipartiteDecisionTreeRegressorSS,
)

from .test_bipartite_trees import compare_trees, assert_equal_leaves
from .utils.make_examples import make_interaction_regression, make_interaction_blobs
from .utils.test_utils import assert_equal_dicts

# from sklearn.tree._tree import DTYPE_t, DOUBLE_t
DTYPE_t, DOUBLE_t = np.float32, np.float64

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("matplotlib").setLevel(logging.CRITICAL)

# FIXME: very small values sometimes cause trees to differ, probably due to
# precision issues.
@pytest.fixture(params=[0.0, 0.0001, 0.1, 0.2328, 0.569, 0.782, 0.995, 1.0])
def supervision(request):
    return request.param


@pytest.fixture(params=range(3))
def random_state(request):
    return request.param


@pytest.fixture(params=[None, 5])
def max_depth(request):
    return request.param

@pytest.fixture
def n_samples():
    return (50, 60)


@pytest.fixture
def n_features():
    return (23, 59)


# NOTE: small leaves may lead to differing trees by chance. Simply becase,
# with just a few samples, one feature column may ocasionally have values with
# the same order as another column, so that they yield the same split position
# and impurities. Since we cannot dictate the order in which features are
# chosen to be evaluated by the splitters, the two equally-ordered features
# can be swapped.
# @pytest.fixture(params=[1, 6, 10, .1])
@pytest.fixture(params=[0.05])
def msl(request):  # min_samples_leaf parameter
    return request.param


@pytest.fixture
def regression_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_targets=None,
        min_target=0.0,
        max_target=1.0,
        noise=0.0,
        return_molten=True,
        return_tree=False,
        random_state=random_state,
        max_depth=None,
    )
    return X, Y, x, y


@pytest.fixture
def classification_data(regression_data):
    X, Y, x, y = regression_data
    Y = (Y > Y.mean()).astype("float64")
    y = (y > y.mean()).astype("float64")
    return X, Y, x, y


@pytest.mark.parametrize(
    "tree",
    [
        DecisionTreeRegressor(),
        DecisionTreeRegressorSS(),
        DecisionTreeRegressorSS(update_supervision=lambda **_: 0.5),
    ],
    ids=["sup", "ss_mono", "ss_bi"],
)
def test_redundant_splits(
    tree,
    supervision,
    random_state,
    classification_data,
):
    X, Y, x, y = classification_data

    tree.set_params(random_state=random_state)
    if not isinstance(tree, DecisionTreeRegressor):
        tree.set_params(supervision=supervision)

    if isinstance(tree, BaseMultipartiteEstimator):
        tree = tree.fit(X, Y)
    else:
        tree = tree.fit(x, y)

    tree_ = tree.tree_
    impurity = tree_.impurity
    single_label_remaining = np.ravel((tree_.value == 0) | (tree_.value == 1))

    # Assert all homogeneous nodes have impurity == 0.
    assert np.allclose(
        impurity[single_label_remaining], 0
    ), "Some homogeneous nodes have impurity != 0."

    # Assert all nodes with impurity == 0 are leaves.
    assert np.all(
        tree_.children_left[impurity == 0]
        == tree_.children_right[impurity == 0]
    ), "Some nodes with impurity == 0 are not leaves."

    # Assert all homogeneous nodes are leaves.
    assert np.all(
        tree_.children_left[single_label_remaining]
        == tree_.children_right[single_label_remaining]
    ), "Some homogeneous nodes are not leaves."


def test_monopartite_semisupervised(
    supervision,
    msl,
    random_state,
    n_samples,
    n_features,
):
    tree_ss = DecisionTreeRegressorSS(
        supervision=supervision,
        min_samples_leaf=msl,
        random_state=random_state,
    )
    tree1 = DecisionTreeRegressor(
        min_samples_leaf=msl,
        random_state=random_state,
    )

    return compare_trees(
        tree1=tree1,
        tree2=tree_ss,
        tree2_is_2d=False,
        supervision=supervision,
        random_state=random_state,
        n_samples=n_samples,
        n_features=n_features,
        min_target=0.0,
        max_target=100.0,
        noise=0.1,
    )


def test_semisupervision_1d2d(
    supervision,
    random_state,
    msl,
    max_depth,
    n_samples,
    n_features,
):
    # n_samples = np.prod(params['n_samples'])
    # max_depth = int(np.ceil(np.log2(n_samples/6))) if msl == 1 else None

    print('* Supervision level:', supervision)
    print('* Max depth:', max_depth)

    tree1 = DecisionTreeRegressorSS(
        supervision=supervision,
        min_samples_leaf=msl,
        random_state=random_state,
        max_depth=max_depth,
    )
    tree2 = BipartiteDecisionTreeRegressorSS(
        supervision=supervision,
        min_samples_leaf=msl,
        random_state=random_state,
        max_depth=max_depth,
    )

    compare_trees(
        tree1=tree1,
        tree2=tree2,
        tree2_is_2d=True,
        supervision=1.0,  # tree1 will already apply the supervision
        random_state=random_state,
        n_samples=n_samples,
        n_features=n_features,
        min_target=0.0,
        max_target=100.0,
        noise=0.1,
    )


@pytest.mark.parametrize('update_supervision', [
    lambda **_: 0.7,
    lambda original_supervision, **_: original_supervision,
    lambda weighted_n_node_samples, weighted_n_samples, **_:
        1 - 0.3 * weighted_n_node_samples/weighted_n_samples,
])
def test_dynamic_supervision_1d2d(
    supervision,
    random_state,
    msl,
    max_depth,
    update_supervision,
    n_samples,
    n_features,
    classification_data,
):
    X, Y, x, y = classification_data

    mono_sup_values = []
    bi_sup_values = []

    def mono_update_supervision(**kwargs):
        new_sup = update_supervision(**kwargs)
        mono_sup_values.append((kwargs['current_supervision'], new_sup))
        return new_sup
    def bi_update_supervision(**kwargs):
        new_sup = update_supervision(**kwargs)
        bi_sup_values.append((kwargs['current_supervision'], new_sup))
        return new_sup

    tree1 = DecisionTreeRegressorSS(
        criterion="squared_error",
        unsupervised_criterion="squared_error",
        supervision=supervision,
        update_supervision=mono_update_supervision,
        min_samples_leaf=msl,
        max_depth=max_depth,
    )
    tree2 = BipartiteDecisionTreeRegressorSS(
        criterion="squared_error",
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        supervision=supervision,
        update_supervision=bi_update_supervision,
        min_samples_leaf=msl,
        max_depth=max_depth,
    )

    tree1.fit(x, y)
    tree2.fit(X, Y)

    # One update at each node.
    assert len(mono_sup_values) == tree1.tree_.node_count
    # One update for each axis at each node.
    assert len(bi_sup_values) == 2 * tree2.tree_.node_count

    # NOTE: the bipartite version will modify its children supervision value
    # before setting a new one.


def test_semisupervision_fixed_sup(
    random_state,
    supervision,
    msl,
    max_depth,
    classification_data,
):
    _, _, x, y = classification_data

    tree1 = DecisionTreeRegressorSS(
        criterion="squared_error",
        unsupervised_criterion="squared_error",
        supervision=supervision,
        update_supervision=None,
        min_samples_leaf=msl,
        max_depth=max_depth,
        random_state=random_state,
    )
    tree2 = clone(tree1).set_params(update_supervision=lambda **_: supervision)

    tree1.fit(x, y)
    tree2.fit(x, y)

    assert_equal_leaves(tree1.tree_, tree2.tree_)




import copy
import logging
import numpy as np
from itertools import product
from pprint import pprint
from typing import Callable

import pytest
from sklearn.tree import DecisionTreeRegressor
from sklearn.base import clone
import sklearn.tree
# from sklearn.tree._tree import Tree
from sklearn.utils._testing import assert_allclose
from sklearn.utils.validation import check_symmetric, check_random_state

from bipartite_learn.tree._bipartite_classes import (
    BipartiteDecisionTreeRegressor,
    BipartiteDecisionTreeClassifier,
    _normalize_weights,
)
from bipartite_learn.tree import _bipartite_classes
from bipartite_learn.melter import row_cartesian_product

from .utils.make_examples import make_interaction_blobs, make_interaction_regression, make_interaction_data
from .utils.test_utils import assert_equal_dicts, stopwatch, parse_args, gen_mock_data, melt_2d_data

# from sklearn.tree._tree import DTYPE_t, DOUBLE_t
DTYPE_t, DOUBLE_t = np.float32, np.float64

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("matplotlib").setLevel(logging.CRITICAL)

# Default test params
DEF_PARAMS = dict(
    n_samples=(51, 60),
    n_features=(10, 9),
    # noise=2,
    noise=0.1,
    # noise=0.0,  # makes it unable to reach 1 sample per leaf
)


@pytest.fixture(params=range(10))
def random_state(request):
    return request.param


@pytest.fixture(params=[(37, 61)])
def n_samples(request):
    return request.param


@pytest.fixture(params=[(10, 9)])
def n_features(request):
    return request.param

@pytest.fixture(params=["squared_error_gso", "friedman_gso"])
def gso_criterion(request):
    return request.param

def get_leaves(tree: sklearn.tree._tree.Tree):
    # the Tree object can be accessed from tree estimators with estimator.tree_
    ch_left = tree.children_left
    ch_right = tree.children_right
    leaf_mask = ch_right == ch_left  # == -1 == sklearn.tree._tree.TREE_LEAF

    return dict(
        value=tree.value[leaf_mask],
        impurity=tree.impurity[leaf_mask],
        n_node_samples=tree.n_node_samples[leaf_mask],
        weighted_n_node_samples=tree.weighted_n_node_samples[leaf_mask],
    )


def assert_equal_leaves(
    tree1, tree2, verbose=False, ignore=None, rtol=1e-7, atol=1e-8,
):
    n_leaves1 = tree1.n_leaves
    n_leaves2 = tree2.n_leaves
    assert n_leaves1 == n_leaves2, (
        f'Number of leaves differ: {n_leaves1} != {n_leaves2}.'
    )
    leaves1 = get_leaves(tree1)
    leaves2 = get_leaves(tree2)

    for leaves in (leaves1, leaves2):
        # It is normal for the last leaves to differ in their order.
        keys_to_order = [
            leaves['impurity'].reshape(-1),
            leaves['weighted_n_node_samples'].reshape(-1),
        ]
        if ignore is None or 'value' not in ignore:
            if leaves1['value'].shape != leaves2['value'].shape:
                raise ValueError(
                    'Diferent leaf value shapes, you are probably working with '
                    'mixed classification-regression trees. To ignore leaf '
                    'values include "value" in the ignore parameter.'
                )
            keys_to_order.append(leaves['value'].reshape(-1))

        order = np.lexsort(np.vstack(keys_to_order))

        for k, v in leaves.items():
            leaves[k] = v[order]
            # leaves[k] = np.sort(leaves[k].flatten())
            # leaves[k] = np.sort(np.unique(leaves[k].flatten()))

    assert_equal_dicts(leaves1, leaves2, ignore=ignore, rtol=rtol, atol=atol)


# TODO: parameter description.
def compare_trees(
    tree1,
    tree2,
    random_state,
    tree2_is_2d=True,
    supervision=1.0,
    max_gen_depth=None,
    verbose=False,
    rtol=1e-7,
    atol=1e-8,
    **params,
):
    """Fit and compare trees.

    Fit bipartite_learnBipartiteDecisionTreeRegressor on mock data and assert the grown
    tree is identical to the one built by sklearn.DecisionTreeRegressor.

    Parameters
    ----------
        random_state : int
        shape : list-like of int
        n_features : list-like of int
        nrules : int
        min_samples_leaf : int
        transpose_test : bool
        noise : float
    """
    print('Starting with settings:')
    pprint(dict(
        random_state=random_state,
        tree2_is_2d=tree2_is_2d,
        supervision=supervision,
        max_gen_depth=max_gen_depth,
    ) | params)

    with stopwatch():
        # XX, Y, x, y, gen_tree = make_interaction_regression(
        #     return_molten=True,
        #     return_tree=True,
        #     n_samples=params['n_samples'],
        #     n_features=params['n_features'],
        #     noise=params['noise'],
        #     random_state=random_state,
        #     max_depth=max_gen_depth,
        #     n_targets=params.get('n_targets'),
        #     max_target=1000,
        # )
        XX, Y, x, y = make_interaction_blobs(
            return_molten=True,
            n_samples=params['n_samples'],
            n_features=params['n_features'],
            random_state=random_state,
            noise=2.0,
            centers=10,
        )

    # NOTE on ExtraTrees:
    # Even with the same random_state, the way 2d splitter uses this random
    # state will be different (same random state for each axis), thus yielding
    # an ExtraTree2d different from sklearn's ExtraTree.

    with stopwatch(f'Fitting {tree1.__class__.__name__}...'):
        if supervision == 1.0:
            print('* Using supervised data for tree1.')
            tree1.fit(x, y)
        elif supervision == 0.0:
            print('* Using unsupervised data for tree1.')
            tree1.fit(x, x)
        else:
            print('* Using semisupervised data for tree1.')
            # Apply 'supervision' parameter weighting
            Xy = np.hstack((x.copy(), y.copy().reshape(-1, 1)))
            # Xy.shape[1] == n_features + n_outputs = n_features + 1
            Xy[:, -1] *= np.sqrt(supervision * Xy.shape[1])
            Xy[:, :-1] *= np.sqrt((1-supervision) *
                                  Xy.shape[1] / (Xy.shape[1]-1))
            tree1.fit(x, Xy)

    with stopwatch(f'Fitting {tree2.__class__.__name__}...'):
        if tree2_is_2d:
            print('* Using 2D data for tree2.')
            tree2.fit(XX, Y)
        else:
            print('* Using 1D data for tree2.')
            tree2.fit(x, y)
    
    if verbose:
        print(f'* Tree 1 ({tree1.__class__.__name__}) params:')
        pprint(tree1.get_params())
        print(f'* Tree 2 ({tree2.__class__.__name__}) params:')
        pprint(tree2.get_params())

    assert_equal_leaves(
        tree1.tree_,
        tree2.tree_,
        ignore=['value'] if supervision != 1.0 else None,
        rtol=rtol,
        atol=atol,
    )


@pytest.mark.parametrize('msl', [1, 5, 100])
@pytest.mark.parametrize('max_depth', [None, 5, 10])
@pytest.mark.parametrize(
    'criterion_mono, criterion_bi',
    [
        ('squared_error', 'squared_error_gso'),
        ('friedman_mse', 'friedman_gso'),
    ]
)
def test_simple_tree_1d2d(
    msl,
    random_state,
    max_depth,
    criterion_mono,
    criterion_bi,
    **params,
):
    params = DEF_PARAMS | params
    tree2 = BipartiteDecisionTreeRegressor(
        min_samples_leaf=msl,
        max_depth=max_depth,
        random_state=check_random_state(random_state),
        bipartite_adapter="gmosa",
        criterion=criterion_bi,
    )

    tree1 = DecisionTreeRegressor(
        min_samples_leaf=msl,
        max_depth=max_depth,
        criterion=criterion_mono,
        random_state=check_random_state(random_state),
    )

    return compare_trees(
        tree1=tree1,
        tree2=tree2,
        random_state=random_state,
        **params,
    )


@pytest.mark.parametrize(
    "pred_weight", [None, "uniform", "precomputed", lambda x: x**2])
class TestGMOSymmetry:
    """Test X symmetry constraints of GMO trees.
    """
    params = DEF_PARAMS
    XX, Y = make_interaction_regression(**params)
    XX_sim, Y_sim, = make_interaction_regression(
        **(params | dict(n_features=params['n_samples'])))

    pbct = BipartiteDecisionTreeRegressor(
        bipartite_adapter="gmo",
    )

    def test_square_array_error(self, pred_weight, **params):
        params = DEF_PARAMS | params
        pbct = clone(self.pbct).set_params(prediction_weights=pred_weight)
        with pytest.raises(
            ValueError,
            match=r"array must be 2-dimensional and square.",
        ):
            pbct.fit(self.XX, self.Y)

    def test_non_symmetric_array(self, pred_weight, **params):
        params = DEF_PARAMS | params
        pbct = clone(self.pbct).set_params(prediction_weights=pred_weight)
        with pytest.raises(
            ValueError,
            match=r"Array must be symmetric",
        ):
            pbct.fit(self.XX_sim, self.Y_sim)

    @pytest.mark.parametrize("splitter", ["random", "best"])
    def test_working(self, pred_weight, splitter, **params):
        params = DEF_PARAMS | params
        pbct = clone(self.pbct).set_params(
            prediction_weights=pred_weight,
            splitter=splitter,
        )
        XX_sim = [check_symmetric(Xi, raise_warning=False)
                  for Xi in self.XX_sim]
        pbct.fit(XX_sim, self.Y_sim)
        pbct.predict(np.hstack([XX_sim[0][:3], XX_sim[1][:3]]))


@pytest.mark.parametrize("splitter", ["random", "best"])
@pytest.mark.parametrize("adapter", ["gmosa"])
def test_identity_gso(splitter, adapter, gso_criterion, **params):
    params = DEF_PARAMS | params
    XX, Y, x, y = make_interaction_regression(return_molten=True, **params)
    tree = BipartiteDecisionTreeRegressor(
        min_samples_leaf=1,
        splitter=splitter,
        bipartite_adapter=adapter,
        criterion=gso_criterion,
    )
    assert_allclose(tree.fit(XX, Y).predict(XX).reshape(Y.shape), Y)


@pytest.mark.parametrize(
    "pred_weights", [
        None,
        *_bipartite_classes.PREDICTION_WEIGHTS_OPTIONS-{"raw"},
        lambda x: x ** 3,
    ]
)
@pytest.mark.parametrize("splitter", ["random", "best"])
@pytest.mark.parametrize(
    "criterion", _bipartite_classes.BIPARTITE_CRITERIA["gmo"]["regression"].keys()
)
def test_identity_gmo(pred_weights, splitter, criterion, **params):
    """Tests wether regression trees can grow util isolating every instance.
    """
    params = DEF_PARAMS | params
    params['n_features'] = params['n_samples'] #= (10, 10)

    XX, Y = make_interaction_regression(**params)
    tree = BipartiteDecisionTreeRegressor(
        min_samples_leaf=1,
        bipartite_adapter="gmo",
        prediction_weights=pred_weights,
        splitter=splitter,
        criterion=criterion,
    )
    XX = [check_symmetric(X, raise_warning=False) for X in XX]
    Y *= 1000

    tree.fit(XX, Y)

    wn_samples = tree.tree_.weighted_n_node_samples
    ch_left = tree.tree_.children_left
    ch_right = tree.tree_.children_right
    n_samples_per_leaf = wn_samples[ch_left == ch_right]

    assert_allclose(n_samples_per_leaf, 1.0)
    assert_allclose(tree.predict(XX).reshape(Y.shape), Y)


@pytest.mark.parametrize("splitter", ["random", "best"])
def test_random_state(splitter, n_samples, random_state):
    random_state = check_random_state(random_state)

    XX = [random_state.uniform(size=(n, n)) for n in n_samples]
    XX = [check_symmetric(X, raise_warning=False) for X in XX]
    Y = random_state.choice((0., 1.), size=n_samples)

    tree1 = BipartiteDecisionTreeRegressor(
        min_samples_leaf=1,
        bipartite_adapter="gmosa",
        splitter=splitter,
        criterion="squared_error",
        random_state=copy.deepcopy(random_state),
    ).fit(XX, Y)

    tree2 = BipartiteDecisionTreeRegressor(
        min_samples_leaf=1,
        bipartite_adapter="gmosa",
        splitter=splitter,
        criterion="squared_error",
        random_state=copy.deepcopy(random_state),
    ).fit(XX, Y)

    leaves1 = get_leaves(tree1.tree_)
    leaves2 = get_leaves(tree2.tree_)

    values1 = tree1.tree_.value[..., -1].reshape(-1)
    values2 = tree2.tree_.value[..., -1].reshape(-1)

    assert len(values1) > 10, "Not enough leaves"

    pred1 = tree1.predict(XX)
    pred2 = tree2.predict(XX)

    assert_equal_dicts(
        {
            **leaves1,
            'value': values1,
            'predictions': pred1,
        },
        {
            **leaves2,
            'value': values2,
            'predictions': pred2,
        },
    )


@pytest.mark.parametrize("splitter", ["random", "best"])
def test_gini_mse_identity(splitter, n_samples, random_state):
    random_state_ = check_random_state(random_state)

    XX = [random_state_.uniform(size=(n, n)) for n in n_samples]
    XX = [check_symmetric(X, raise_warning=False) for X in XX]
    Y = random_state_.choice((0., 1.), size=n_samples)

    # NOTE on random_state: passing the integer itself to the trees was the only
    # option to yield consistent results. Even deepcopying the random_state
    # object seems to induce some differences.
    tree_reg = BipartiteDecisionTreeRegressor(
        min_rows_leaf=10,
        min_cols_leaf=10,
        bipartite_adapter="gmosa",
        splitter=splitter,
        criterion="squared_error",
        random_state=random_state,
    ).fit(XX, Y)

    tree_clf = BipartiteDecisionTreeClassifier(
        min_rows_leaf=10,
        min_cols_leaf=10,
        bipartite_adapter="gmosa",
        splitter=splitter,
        criterion="gini",
        random_state=random_state,
    ).fit(XX, Y)

    leaves_reg = get_leaves(tree_reg.tree_)
    leaves_clf = get_leaves(tree_clf.tree_)

    leaves_reg['impurity'] *= 2

    values_reg = tree_reg.tree_.value[..., -1].reshape(-1)
    values_clf = tree_clf.tree_.value[..., -1].flatten()  # copy
    values_clf /= tree_clf.tree_.weighted_n_node_samples

    assert len(leaves_reg['impurity']) > 5, "Not enough leaves"

    pred_reg = tree_reg.predict(XX)
    pred_clf = tree_clf.predict_proba(XX)[..., -1]

    assert_equal_dicts(
        {
            **leaves_reg,
            'value': values_reg,
            'predictions': pred_reg,
        },
        {
            **leaves_clf,
            'value': values_clf,
            'predictions': pred_clf,
        },
    )


@pytest.mark.parametrize("min_samples_leaf", [1, 20, 100])
@pytest.mark.parametrize("splitter", ["random", "best"])
@pytest.mark.parametrize(
    "criterion", _bipartite_classes.BIPARTITE_CRITERIA["gmo"]["regression"].keys()
)
def test_leaf_mean_symmetry(min_samples_leaf, splitter, criterion):
    """Tests if Y_leaf.mean(0).mean() == Y_leaf.mean(1).mean()
    
    Y_leaf.mean(0) and Y_leaf.mean(1) are stored by regression GMO trees as
    output values. This function tests if these values are correct by checking
    the aforementioned equality.
    """
    params = DEF_PARAMS

    XX, Y, x, y = make_interaction_regression(return_molten=True, **params)
    tree = BipartiteDecisionTreeRegressor(
        min_samples_leaf=min_samples_leaf,
        bipartite_adapter="gmo",
        prediction_weights="raw",
        splitter=splitter,
        criterion=criterion,
    )
    tree.fit(XX, Y)
    pred = tree.predict(XX)
    row_pred, col_pred = np.hsplit(pred, [tree._n_rows_fit])
    mean_rows = np.nanmean(row_pred, axis=1)
    mean_cols = np.nanmean(col_pred, axis=1)

    assert not np.isnan(mean_rows).any()
    assert not np.isnan(mean_cols).any()
    assert_allclose(mean_rows, mean_cols)


def test_weight_normalization():
    rng = np.random.default_rng()
    w = rng.random((100, 20))
    zeroed_w = rng.choice(w.shape[0], w.shape[0]//10)
    w[zeroed_w] = 0.

    pred = rng.random((100, 20))
    not_nan = rng.choice((True, False), pred.shape)
    pred[~not_nan] = np.nan

    w = _normalize_weights(w, pred)

    assert_allclose(w.sum(axis=1), 1.)
    assert all(
        1. not in row or set(row.unique()) < {0., 1.}
        for row in w
    )
    assert np.all(w[np.isnan(pred)] == 0.)
    assert_allclose(
        w[zeroed_w],
        (not_nan / not_nan.sum(axis=1, keepdims=True))[zeroed_w]
    )


@pytest.mark.parametrize("mrl,mcl", [(1, 5), (2, 1), (11, 19)])
@pytest.mark.parametrize("criterion", ["squared_error"])
def test_leaf_shape_gmo(random_state, mrl, mcl, criterion, **params):
    params = DEF_PARAMS | params
    # XX, Y, x, y = gen_mock_data(melt=True, **params)
    rng = check_random_state(random_state)
    n_clusters = 10, 10
    Y_unique = np.arange(np.prod(n_clusters), dtype=float).reshape(n_clusters)
    Y = Y_unique.repeat(mrl, axis=0).repeat(mcl, axis=1)
    shape = mrl * n_clusters[0], mcl * n_clusters[1]

    n_features = 50, 50
    XX = [
        rng.random((s, f)).astype(np.float32)
        for s, f in zip(shape, n_features)
    ]
    XX[0][:, 0] = np.sort(XX[0][:, 0])
    XX[1][:, 0] = np.sort(XX[1][:, 0])

    # Shuffle
    id_rows = rng.choice(shape[0], size=shape[0], replace=False)
    id_cols = rng.choice(shape[1], size=shape[1], replace=False)
    XX[0] = XX[0][id_rows]
    XX[1] = XX[1][id_cols]
    Y = Y[np.ix_(id_rows, id_cols)]

    tree = BipartiteDecisionTreeRegressor(
        min_rows_leaf=mrl,
        min_cols_leaf=mcl,
        bipartite_adapter="gmo",
        prediction_weights="raw",
        criterion=criterion,
    )
    tree.fit(XX, Y)
    pred = tree.predict(XX)
    row_pred, col_pred = np.hsplit(pred, [tree._n_rows_fit])
    n_rows = (~np.isnan(row_pred)).sum(axis=1)
    n_cols = (~np.isnan(col_pred)).sum(axis=1)

    assert_allclose(n_rows, mrl)
    assert_allclose(n_cols, mcl)


@pytest.mark.parametrize("mrl,mcl", [(1, 1), (1, 5), (2, 1), (19, 11)])
def test_leaf_shape_gso(mrl, mcl, random_state, gso_criterion, **params):
    params = DEF_PARAMS | params
    # XX, Y, x, y = gen_mock_data(melt=True, **params)
    rng = check_random_state(random_state)
    n_clusters = 10, 10
    Y_unique = np.arange(np.prod(n_clusters), dtype=float).reshape(n_clusters)
    Y = Y_unique.repeat(mrl, axis=0).repeat(mcl, axis=1)
    y_values = np.sort(Y_unique.reshape(-1))

    shape = mrl * n_clusters[0], mcl * n_clusters[1]
    n_features = 50, 50
    XX = [
        rng.random((s, f)).astype(np.float32)
        for s, f in zip(shape, n_features)
    ]
    XX[0][:, 0] = np.sort(XX[0][:, 0])
    XX[1][:, 0] = np.sort(XX[1][:, 0])

    # Shuffle
    id_rows = rng.choice(shape[0], size=shape[0], replace=False)
    id_cols = rng.choice(shape[1], size=shape[1], replace=False)
    XX[0] = XX[0][id_rows]
    XX[1] = XX[1][id_cols]
    Y = Y[np.ix_(id_rows, id_cols)]

    # Control result
    X, y = row_cartesian_product(XX), Y.reshape(-1)
    tree1d = DecisionTreeRegressor()
    tree1d.fit(X, y)

    tree = BipartiteDecisionTreeRegressor(
        min_rows_leaf=mrl,
        min_cols_leaf=mcl,
        bipartite_adapter="gmosa",
        criterion=gso_criterion,
    )
    tree.fit(XX, Y)

    leaves1 = get_leaves(tree1d.tree_)
    leaf_values1d = np.sort(leaves1['value'].reshape(-1))
    assert_allclose(leaves1['n_node_samples'], mrl*mcl)
    assert_allclose(leaf_values1d, y_values)

    assert_equal_leaves(tree1d.tree_, tree.tree_)
